#include <torch/extension.h>
#include <ATen/ATen.h>
#include <ATen/cuda/CUDAContext.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>
#include <math.h>

#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) CHECK_CUDA(x); CHECK_CONTIGUOUS(x)

__device__ __forceinline__ float4 load_float4_aligned(const float* addr) {
    return *reinterpret_cast<const float4*>(addr);
}

__global__ void efficient_fused_kernel(
    const float* __restrict__ input,
    float* __restrict__ output,
    const float* __restrict__ gamma,
    const float* __restrict__ beta,
    int N, int C, int D, int H, int W,
    int groups, float eps
) {
    int n = blockIdx.x;
    int g = blockIdx.y;

    int channels_per_group = C / groups;
    int group_elements = channels_per_group * D * H * W;
    int base = n * (C * D * H * W) + g * group_elements;

    int tid = threadIdx.x;
    int blockSize = blockDim.x;

    float local_sum = 0.0f;
    float local_sumsq = 0.0f;

    for (int i = tid; i < group_elements; i += blockSize) {
        int index = base + i;
        float x = input[index];
        float sw = x / (1.0f + expf(-x));
        local_sum += sw;
        local_sumsq += sw * sw;
    }

    unsigned int mask = 0xffffffff;
    for (int offset = warpSize / 2; offset > 0; offset /= 2) {
        local_sum += __shfl_down_sync(mask, local_sum, offset);
        local_sumsq += __shfl_down_sync(mask, local_sumsq, offset);
    }

    int warpId = tid / warpSize;
    int laneId = tid % warpSize;

    extern __shared__ float shared[];
    float* warp_sums = shared;
    float* warp_sumsq = shared + blockDim.x/warpSize;

    if (laneId == 0) {
        warp_sums[warpId] = local_sum;
        warp_sumsq[warpId] = local_sumsq;
    }
    __syncthreads();

    if (tid < (blockDim.x / warpSize)) {
        local_sum = warp_sums[tid];
        local_sumsq = warp_sumsq[tid];
    }

    if (tid == 0) {
        float total_sum = 0.0f;
        float total_sumsq = 0.0f;
        for (int i = 0; i < blockDim.x/warpSize; i++) {
            total_sum += warp_sums[i];
            total_sumsq += warp_sumsq[i];
        }
        warp_sums[0] = total_sum;
        warp_sumsq[0] = total_sumsq;
    }
    __syncthreads();

    float mean = warp_sums[0] / group_elements;
    float var = warp_sumsq[0] / group_elements - mean * mean;
    float inv_std = rsqrtf(var + eps);

    for (int i = tid; i < group_elements; i += blockSize) {
        int index = base + i;
        float x = input[index];
        float sw = x / (1.0f + expf(-x));
        int local_channel = i / (D * H * W);
        int global_channel = g * channels_per_group + local_channel;

        float norm = (sw - mean) * inv_std;
        float y = norm * gamma[global_channel] + beta[global_channel];
        float hs = y * fminf(fmaxf(y + 3.0f, 0.0f), 6.0f) / 6.0f;
        output[index] = hs;
    }
}

torch::Tensor forward(
    torch::Tensor x,
    int stride,
    int padding,
    int groups,
    float eps,
    torch::Tensor conv_transpose,
    torch::Tensor conv_transpose_bias,
    torch::Tensor group_norm_weight,
    torch::Tensor group_norm_bias
) {
    CHECK_INPUT(x);
    CHECK_INPUT(conv_transpose);
    CHECK_INPUT(conv_transpose_bias);
    CHECK_INPUT(group_norm_weight);
    CHECK_INPUT(group_norm_bias);

    x = torch::conv_transpose3d(x, conv_transpose, conv_transpose_bias, stride, padding);
    torch::Tensor output = torch::empty_like(x);

    int N = x.size(0);
    int C = x.size(1);
    int D = x.size(2);
    int H = x.size(3);
    int W = x.size(4);

    dim3 grid(N, groups);
    int threads = 256;
    size_t shared_mem_size = 2 * (threads/32) * sizeof(float);

    efficient_fused_kernel<<<grid, threads, shared_mem_size>>>(
        x.data_ptr<float>(),
        output.data_ptr<float>(),
        group_norm_weight.data_ptr<float>(),
        group_norm_bias.data_ptr<float>(),
        N, C, D, H, W,
        groups,
        eps
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Efficient fused kernel");
}